@@ -103,19 +103,19 @@ module Nonexhaustive =
   end
 ;;
 [%%expect{|
-Line _, characters 6-34:
-  ......function
-          | C2 x -> x
-Warning 8: this pattern-matching is not exhaustive.
+Lines 11-12, characters 6-19:
+11 | ......function
+12 |         | C2 x -> x
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 C1 _
-Line _, characters 6-77:
-  ......function
-          | Foo _ , Foo _ -> true
-          | Bar _, Bar _ -> true
-Warning 8: this pattern-matching is not exhaustive.
+Lines 24-26, characters 6-30:
+24 | ......function
+25 |         | Foo _ , Foo _ -> true
+26 |         | Bar _, Bar _ -> true
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
-(Bar _, Foo _)
+(Foo _, Bar _)
 module Nonexhaustive :
   sig
     type 'a u = C1 : int -> int u | C2 : bool -> bool u
@@ -157,16 +157,16 @@ module PR6862 = struct
   class d (Just x) = object method x : int = x end
 end;;
 [%%expect{|
-Line _, characters 10-18:
-    class c (Some x) = object method x : int = x end
-            ^^^^^^^^
-Warning 8: this pattern-matching is not exhaustive.
+Line 2, characters 10-18:
+2 |   class c (Some x) = object method x : int = x end
+              ^^^^^^^^
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 None
-Line _, characters 10-18:
-    class d (Just x) = object method x : int = x end
-            ^^^^^^^^
-Warning 8: this pattern-matching is not exhaustive.
+Line 4, characters 10-18:
+4 |   class d (Just x) = object method x : int = x end
+              ^^^^^^^^
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 Nothing
 module PR6862 :
@@ -192,10 +192,10 @@ module PR6220 = struct
   let g : int t -> int = function I -> 1 | _ -> 2 (* warn *)
 end;;
 [%%expect{|
-Line _, characters 43-44:
-    let g : int t -> int = function I -> 1 | _ -> 2 (* warn *)
-                                             ^
-Warning 56: this match case is unreachable.
+Line 4, characters 43-44:
+4 |   let g : int t -> int = function I -> 1 | _ -> 2 (* warn *)
+                                               ^
+Warning 56 [unreachable-case]: this match case is unreachable.
 Consider replacing it with a refutation case '<pat> -> .'
 module PR6220 :
   sig
@@ -260,10 +260,10 @@ module PR6801 = struct
     | String s -> print_endline s (* warn : Any *)
 end;;
 [%%expect{|
-Line _, characters 4-50:
-  ....match x with
-      | String s -> print_endline s.................
-Warning 8: this pattern-matching is not exhaustive.
+Lines 8-9, characters 4-33:
+8 | ....match x with
+9 |     | String s -> print_endline s.................
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 Any
 module PR6801 :
@@ -284,9 +284,9 @@ module Existential_escape =
   end
 ;;
 [%%expect{|
-Line _, characters 21-22:
-      let eval (D x) = x
-                       ^
+Line 5, characters 21-22:
+5 |     let eval (D x) = x
+                         ^
 Error: This expression has type $D_'a t
        but an expression was expected of type 'a
        The type constructor $D_'a would escape its scope
@@ -317,12 +317,11 @@ module Or_patterns =
 end
 ;;
 [%%expect{|
-Line _, characters 11-19:
-          | (IntLit _ | BoolLit _) -> ()
-             ^^^^^^^^
-Error: This pattern matches values of type int t
-       but a pattern was expected which matches values of type s t
-       Type int is not compatible with type s
+module Or_patterns :
+  sig
+    type _ t = IntLit : int -> int t | BoolLit : bool -> bool t
+    val eval : 's t -> unit
+  end
 |}];;
 
 module Polymorphic_variants =
@@ -368,9 +367,9 @@ module Propagation :
     val check : 's t -> 's
   end
 |}, Principal{|
-Line _, characters 19-20:
-      | BoolLit b -> b
-                     ^
+Line 13, characters 19-20:
+13 |     | BoolLit b -> b
+                        ^
 Error: This expression has type bool but an expression was expected of type s
 |}];;
 
@@ -381,17 +380,11 @@ module Normal_constrs = struct
   let f = function A -> 1 | B -> 2
 end;;
 [%%expect{|
-Line _, characters 28-29:
-    let f = function A -> 1 | B -> 2
-                              ^
+Line 5, characters 28-29:
+5 |   let f = function A -> 1 | B -> 2
+                                ^
 Error: This variant pattern is expected to have type a
        The constructor B does not belong to type a
-|}, Principal{|
-Line _, characters 28-29:
-    let f = function A -> 1 | B -> 2
-                              ^
-Error: This pattern matches values of type b
-       but a pattern was expected which matches values of type a
 |}];;
 
 module PR6849 = struct
@@ -401,9 +394,9 @@ module PR6849 = struct
       Foo -> 5
 end;;
 [%%expect{|
-Line _, characters 6-9:
-        Foo -> 5
-        ^^^
+Line 5, characters 6-9:
+5 |       Foo -> 5
+          ^^^
 Error: This pattern matches values of type 'a t
        but a pattern was expected which matches values of type int
 |}];;
@@ -432,9 +425,9 @@ let test : type a. a t -> _ =
   function Int -> ky (1 : a) 1  (* fails *)
 ;;
 [%%expect{|
-Line _, characters 18-30:
-    function Int -> ky (1 : a) 1  (* fails *)
-                    ^^^^^^^^^^^^
+Line 2, characters 18-30:
+2 |   function Int -> ky (1 : a) 1  (* fails *)
+                      ^^^^^^^^^^^^
 Error: This expression has type a = int
        but an expression was expected of type 'a
        This instance of int is ambiguous:
@@ -446,9 +439,9 @@ let test : type a. a t -> a = fun x ->
   in r
 ;;
 [%%expect{|
-Line _, characters 30-42:
-    let r = match x with Int -> ky (1 : a) 1  (* fails *)
-                                ^^^^^^^^^^^^
+Line 2, characters 30-42:
+2 |   let r = match x with Int -> ky (1 : a) 1  (* fails *)
+                                  ^^^^^^^^^^^^
 Error: This expression has type a = int
        but an expression was expected of type 'a
        This instance of int is ambiguous:
@@ -460,9 +453,9 @@ let test : type a. a t -> a = fun x ->
   in r
 ;;
 [%%expect{|
-Line _, characters 30-42:
-    let r = match x with Int -> ky 1 (1 : a)  (* fails *)
-                                ^^^^^^^^^^^^
+Line 2, characters 30-42:
+2 |   let r = match x with Int -> ky 1 (1 : a)  (* fails *)
+                                  ^^^^^^^^^^^^
 Error: This expression has type a = int
        but an expression was expected of type 'a
        This instance of int is ambiguous:
@@ -536,9 +529,9 @@ let test2 : type a. a t -> a option = fun x ->
   !u
 ;; (* fails because u : (int | a) option ref *)
 [%%expect{|
-Line _, characters 46-48:
-    begin match x with Int -> u := Some 1; r := !u end;
-                                                ^^
+Line 4, characters 46-48:
+4 |   begin match x with Int -> u := Some 1; r := !u end;
+                                                  ^^
 Error: This expression has type int option
        but an expression was expected of type a option
        Type int is not compatible with type a = int
@@ -574,9 +567,9 @@ let we_y1x (type a) (x : a) (v : a t) =
 ;; (* fail *)
 [%%expect{|
 val either : 'a -> 'a -> 'a = <fun>
-Line _, characters 44-45:
-    match v with Int -> let y = either 1 x in y
-                                              ^
+Line 3, characters 44-45:
+3 |   match v with Int -> let y = either 1 x in y
+                                                ^
 Error: This expression has type a = int
        but an expression was expected of type 'a
        This instance of int is ambiguous:
@@ -635,9 +628,9 @@ let f (type a) (x : a t) y =
     in M.z
 ;; (* fails because of aliasing... *)
 [%%expect{|
-Line _, characters 46-47:
-      let module M = struct type b = a let z = (y : b) end
-                                                ^
+Line 3, characters 46-47:
+3 |     let module M = struct type b = a let z = (y : b) end
+                                                  ^
 Error: This expression has type a = int
        but an expression was expected of type b = int
        This instance of int is ambiguous:
@@ -688,9 +681,9 @@ let f : type a b. (a,b) eq -> (<m : a; ..> as 'c) -> (<m : b; ..> as 'c) =
 ;; (* fail *)
 [%%expect{|
 type (_, _) eq = Eq : ('a, 'a) eq
-Line _, characters 4-90:
-  ....f : type a b. (a,b) eq -> (<m : a; ..> as 'c) -> (<m : b; ..> as 'c) =
-    fun Eq o -> o
+Lines 3-4, characters 4-15:
+3 | ....f : type a b. (a,b) eq -> (<m : a; ..> as 'c) -> (<m : b; ..> as 'c) =
+4 |   fun Eq o -> o
 Error: The universal type variable 'b cannot be generalized:
        it is already bound to another variable.
 |}];;
@@ -699,9 +692,9 @@ let f : type a b. (a,b) eq -> <m : a; ..> -> <m : b; ..> =
   fun Eq o -> o
 ;; (* fail *)
 [%%expect{|
-Line _, characters 14-15:
-    fun Eq o -> o
-                ^
+Line 2, characters 14-15:
+2 |   fun Eq o -> o
+                  ^
 Error: This expression has type < m : a; .. >
        but an expression was expected of type < m : b; .. >
        Type a is not compatible with type b = a
@@ -712,9 +705,9 @@ Error: This expression has type < m : a; .. >
 let f (type a) (type b) (eq : (a,b) eq) (o : <m : a; ..>) : <m : b; ..> =
   match eq with Eq -> o ;; (* should fail *)
 [%%expect{|
-Line _, characters 22-23:
-    match eq with Eq -> o ;; (* should fail *)
-                        ^
+Line 2, characters 22-23:
+2 |   match eq with Eq -> o ;; (* should fail *)
+                          ^
 Error: This expression has type < m : a; .. >
        but an expression was expected of type < m : b; .. >
        Type a is not compatible with type b = a
@@ -752,9 +745,9 @@ let f : type a b. (a,b) eq -> < m : a; .. > -> < m : b > =
 [%%expect{|
 val f : ('a, 'b) eq -> < m : 'a > -> < m : 'b > = <fun>
 |}, Principal{|
-Line _, characters 44-45:
-      let r : < m : b > = match eq with Eq -> o in (* fail with principal *)
-                                              ^
+Line 4, characters 44-45:
+4 |     let r : < m : b > = match eq with Eq -> o in (* fail with principal *)
+                                                ^
 Error: This expression has type < m : a >
        but an expression was expected of type < m : b >
        Type a is not compatible with type b = a
@@ -768,9 +761,9 @@ let f : type a b. (a,b) eq -> < m : a; .. > -> < m : b > =
     ignore (o : < m : a >);
     r;;
 [%%expect{|
-Line _, characters 44-45:
-      let r : < m : b > = match eq with Eq -> o in (* fail *)
-                                              ^
+Line 3, characters 44-45:
+3 |     let r : < m : b > = match eq with Eq -> o in (* fail *)
+                                                ^
 Error: This expression has type < m : a; .. >
        but an expression was expected of type < m : b >
        Type a is not compatible with type b = a
@@ -781,18 +774,18 @@ Error: This expression has type < m : a; .. >
 let f : type a b. (a,b) eq -> [> `A of a] -> [> `A of b] =
   fun Eq o -> o ;; (* fail *)
 [%%expect{|
-Line _, characters 14-15:
-    fun Eq o -> o ;; (* fail *)
-                ^
+Line 2, characters 14-15:
+2 |   fun Eq o -> o ;; (* fail *)
+                  ^
 Error: This expression has type [> `A of a ]
        but an expression was expected of type [> `A of b ]
        Type a is not compatible with type b = a
        This instance of a is ambiguous:
        it would escape the scope of its equation
 |}, Principal{|
-Line _, characters 9-15:
-    fun Eq o -> o ;; (* fail *)
-           ^^^^^^
+Line 2, characters 9-15:
+2 |   fun Eq o -> o ;; (* fail *)
+             ^^^^^^
 Error: This expression has type ([> `A of b ] as 'a) -> 'a
        but an expression was expected of type [> `A of a ] -> [> `A of b ]
        Types for tag `A are incompatible
@@ -801,9 +794,9 @@ Error: This expression has type ([> `A of b ] as 'a) -> 'a
 let f (type a b) (eq : (a,b) eq) (v : [> `A of a]) : [> `A of b] =
   match eq with Eq -> v ;; (* should fail *)
 [%%expect{|
-Line _, characters 22-23:
-    match eq with Eq -> v ;; (* should fail *)
-                        ^
+Line 2, characters 22-23:
+2 |   match eq with Eq -> v ;; (* should fail *)
+                          ^
 Error: This expression has type [> `A of a ]
        but an expression was expected of type [> `A of b ]
        Type a is not compatible with type b = a
@@ -814,12 +807,13 @@ Error: This expression has type [> `A of a ]
 let f : type a b. (a,b) eq -> [< `A of a | `B] -> [< `A of b | `B] =
   fun Eq o -> o ;; (* fail *)
 [%%expect{|
-Line _, characters 4-84:
-  ....f : type a b. (a,b) eq -> [< `A of a | `B] -> [< `A of b | `B] =
-    fun Eq o -> o..............
+Lines 1-2, characters 4-15:
+1 | ....f : type a b. (a,b) eq -> [< `A of a | `B] -> [< `A of b | `B] =
+2 |   fun Eq o -> o..............
 Error: This definition has type
-         ('a, 'b) eq -> ([< `A of 'b & 'a | `B ] as 'c) -> 'c
-       which is less general than 'a0 'b0. ('a0, 'b0) eq -> 'c -> 'c
+         'c. ('d, 'c) eq -> ([< `A of 'c & 'f & 'd | `B ] as 'e) -> 'e
+       which is less general than
+         'a 'b. ('a, 'b) eq -> ([< `A of 'b & 'h | `B ] as 'g) -> 'g
 |}];;
 
 let f : type a b. (a,b) eq -> [`A of a | `B] -> [`A of b | `B] =
@@ -843,9 +837,9 @@ let f : type a b. (a,b) eq -> [> `A of a | `B] -> [`A of b | `B] =
 [%%expect{|
 val f : ('a, 'b) eq -> [ `A of 'a | `B ] -> [ `A of 'b | `B ] = <fun>
 |}, Principal{|
-Line _, characters 49-50:
-      let r : [`A of b | `B] = match eq with Eq -> o in (* fail with principal *)
-                                                   ^
+Line 4, characters 49-50:
+4 |     let r : [`A of b | `B] = match eq with Eq -> o in (* fail with principal *)
+                                                     ^
 Error: This expression has type [ `A of a | `B ]
        but an expression was expected of type [ `A of b | `B ]
        Type a is not compatible with type b = a
@@ -859,9 +853,9 @@ let f : type a b. (a,b) eq -> [> `A of a | `B] -> [`A of b | `B] =
     ignore (o : [< `A of a | `B]);
     r;;
 [%%expect{|
-Line _, characters 49-50:
-      let r : [`A of b | `B] = match eq with Eq -> o in (* fail *)
-                                                   ^
+Line 3, characters 49-50:
+3 |     let r : [`A of b | `B] = match eq with Eq -> o in (* fail *)
+                                                     ^
 Error: This expression has type [> `A of a | `B ]
        but an expression was expected of type [ `A of b | `B ]
        Type a is not compatible with type b = a
@@ -916,15 +910,15 @@ let f : type a. a ty -> a t -> int = fun x y ->
   | TA, D z -> z
 ;; (* warn *)
 [%%expect{|
-Line _, characters 2-153:
-  ..match x, y with
-    | _, A z -> z
-    | _, B z -> if z then 1 else 2
-    | _, C z -> truncate z
-    | TE TC, D [|1.0|] -> 14
-    | TA, D 0 -> -1
-    | TA, D z -> z
-Warning 8: this pattern-matching is not exhaustive.
+Lines 2-8, characters 2-16:
+2 | ..match x, y with
+3 |   | _, A z -> z
+4 |   | _, B z -> if z then 1 else 2
+5 |   | _, C z -> truncate z
+6 |   | TE TC, D [|1.0|] -> 14
+7 |   | TA, D 0 -> -1
+8 |   | TA, D z -> z
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 (TE TC, D [| 0. |])
 val f : 'a ty -> 'a t -> int = <fun>
@@ -940,9 +934,9 @@ let f : type a. a ty -> a t -> int = fun x y ->
   | D z, TA -> z
 ;; (* fail *)
 [%%expect{|
-Line _, characters 6-13:
-    | D [|1.0|], TE TC -> 14
-        ^^^^^^^
+Line 6, characters 6-13:
+6 |   | D [|1.0|], TE TC -> 14
+          ^^^^^^^
 Error: This pattern matches values of type 'a array
        but a pattern was expected which matches values of type a
 |}];;
@@ -960,9 +954,9 @@ let f : type a. a ty -> a t -> int = fun x y ->
 ;; (* fail *)
 [%%expect{|
 type ('a, 'b) pair = { right : 'a; left : 'b; }
-Line _, characters 25-32:
-    | {left=TE TC; right=D [|1.0|]} -> 14
-                           ^^^^^^^
+Line 8, characters 25-32:
+8 |   | {left=TE TC; right=D [|1.0|]} -> 14
+                             ^^^^^^^
 Error: This pattern matches values of type 'a array
        but a pattern was expected which matches values of type a
 |}];;
@@ -980,15 +974,15 @@ let f : type a. a ty -> a t -> int = fun x y ->
 ;; (* ok *)
 [%%expect{|
 type ('a, 'b) pair = { left : 'a; right : 'b; }
-Line _, characters 2-244:
-  ..match {left=x; right=y} with
-    | {left=_; right=A z} -> z
-    | {left=_; right=B z} -> if z then 1 else 2
-    | {left=_; right=C z} -> truncate z
-    | {left=TE TC; right=D [|1.0|]} -> 14
-    | {left=TA; right=D 0} -> -1
-    | {left=TA; right=D z} -> z
-Warning 8: this pattern-matching is not exhaustive.
+Lines 4-10, characters 2-29:
+ 4 | ..match {left=x; right=y} with
+ 5 |   | {left=_; right=A z} -> z
+ 6 |   | {left=_; right=B z} -> if z then 1 else 2
+ 7 |   | {left=_; right=C z} -> truncate z
+ 8 |   | {left=TE TC; right=D [|1.0|]} -> 14
+ 9 |   | {left=TA; right=D 0} -> -1
+10 |   | {left=TA; right=D z} -> z
+Warning 8 [partial-match]: this pattern-matching is not exhaustive.
 Here is an example of a case that is not matched:
 {left=TE TC; right=D [| 0. |]}
 val f : 'a ty -> 'a t -> int = <fun>
@@ -1005,9 +999,9 @@ let f : type a b. (a M.t, b M.t) eq -> (a, b) eq =
 ;;
 [%%expect{|
 module M : sig type 'a t val eq : ('a t, 'b t) eq end
-Line _, characters 17-19:
-    function Eq -> Eq (* fail *)
-                   ^^
+Line 6, characters 17-19:
+6 |   function Eq -> Eq (* fail *)
+                     ^^
 Error: This expression has type (a, a) eq
        but an expression was expected of type (a, b) eq
        Type a is not compatible with type b
@@ -1061,9 +1055,9 @@ let g (type t) (x:t) (e : t int_foo) (e' : t int_bar) =
 [%%expect{|
 type _ int_foo = IF_constr : < foo : int; .. > int_foo
 type _ int_bar = IB_constr : < bar : int; .. > int_bar
-Line _, characters 3-4:
-    (x:<foo:int>)
-     ^
+Line 10, characters 3-4:
+10 |   (x:<foo:int>)
+        ^
 Error: This expression has type t = < foo : int; .. >
        but an expression was expected of type < foo : int >
        Type $0 = < bar : int; .. > is not compatible with type <  >
@@ -1075,9 +1069,9 @@ let g (type t) (x:t) (e : t int_foo) (e' : t int_bar) =
   (x:<foo:int;bar:int>)
 ;;
 [%%expect{|
-Line _, characters 3-4:
-    (x:<foo:int;bar:int>)
-     ^
+Line 3, characters 3-4:
+3 |   (x:<foo:int;bar:int>)
+       ^
 Error: This expression has type t = < foo : int; .. >
        but an expression was expected of type < bar : int; foo : int >
        Type $0 = < bar : int; .. > is not compatible with type < bar : int >
@@ -1089,9 +1083,9 @@ let g (type t) (x:t) (e : t int_foo) (e' : t int_bar) =
   (x:<foo:int;bar:int;..>)
 ;;
 [%%expect{|
-Line _, characters 2-26:
-    (x:<foo:int;bar:int;..>)
-    ^^^^^^^^^^^^^^^^^^^^^^^^
+Line 3, characters 2-26:
+3 |   (x:<foo:int;bar:int;..>)
+      ^^^^^^^^^^^^^^^^^^^^^^^^
 Error: This expression has type < bar : int; foo : int; .. >
        but an expression was expected of type 'a
        The type constructor $1 would escape its scope
@@ -1102,7 +1096,7 @@ let g (type t) (x:t) (e : t int_foo) (e' : t int_bar) : t =
   (x:<foo:int;bar:int;..>)
 ;;
 [%%expect{|
-val g : 'a -> 'a int_foo -> 'a int_bar -> 'a = <fun>
+val g : 't -> 't int_foo -> 't int_bar -> 't = <fun>
 |}];;
 
 let g (type t) (x:t) (e : t int_foo) (e' : t int_bar) =
@@ -1110,7 +1104,7 @@ let g (type t) (x:t) (e : t int_foo) (e' : t int_bar) =
   x, x#foo, x#bar
 ;;
 [%%expect{|
-val g : 'a -> 'a int_foo -> 'a int_bar -> 'a * int * int = <fun>
+val g : 't -> 't int_foo -> 't int_bar -> 't * int * int = <fun>
 |}];;
 
 (* PR#5554 *)
@@ -1173,9 +1167,9 @@ let f : type a b. (a,b) eq -> (a,int) eq -> a -> b -> _ = fun ab aint a b ->
   in ignore x
 ;; (* ok *)
 [%%expect{|
-Line _, characters 24-25:
-      if true then a else b
-                          ^
+Line 5, characters 24-25:
+5 |     if true then a else b
+                            ^
 Error: This expression has type b = int
        but an expression was expected of type a = int
        Type b = int is not compatible with type int
@@ -1191,9 +1185,9 @@ let f : type a b. (a,b) eq -> (b,int) eq -> a -> b -> _ = fun ab bint a b ->
   in ignore x
 ;; (* ok *)
 [%%expect{|
-Line _, characters 24-25:
-      if true then a else b
-                          ^
+Line 5, characters 24-25:
+5 |     if true then a else b
+                            ^
 Error: This expression has type b = int
        but an expression was expected of type a = int
        Type int is not compatible with type a = int
@@ -1207,9 +1201,9 @@ let f (type a b c) (b : bool) (w1 : (a,b) eq) (w2 : (a,int) eq) (x : a) (y : b)
   if b then x else y
 ;;
 [%%expect{|
-Line _, characters 19-20:
-    if b then x else y
-                     ^
+Line 4, characters 19-20:
+4 |   if b then x else y
+                       ^
 Error: This expression has type b = int
        but an expression was expected of type a = int
        Type a = int is not compatible with type a = int
@@ -1222,9 +1216,9 @@ let f (type a b c) (b : bool) (w1 : (a,b) eq) (w2 : (a,int) eq) (x : a) (y : b)
   let Eq = w2 in
   if b then y else x
 [%%expect{|
-Line _, characters 19-20:
-    if b then y else x
-                     ^
+Line 4, characters 19-20:
+4 |   if b then y else x
+                       ^
 Error: This expression has type a = int
        but an expression was expected of type b = int
        This instance of int is ambiguous:
